from monk.gluon.losses.imports import *
from monk.system.imports import *


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def load_loss(system_dict):
	'''
    Load loss function in native library

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
	name = system_dict["local"]["criterion"];

	if(name == "l1"):
		system_dict["local"]["criterion"] = mx.gluon.loss.L1Loss(
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"]);

	elif(name == "l2"):
		system_dict["local"]["criterion"] = mx.gluon.loss.L2Loss(
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"]);

	elif(name == "softmaxcrossentropy"):
		system_dict["local"]["criterion"] = mx.gluon.loss.SoftmaxCrossEntropyLoss(
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"],
			axis=system_dict["hyper-parameters"]["loss"]["params"]["axis_to_sum_over"],
			sparse_label=system_dict["hyper-parameters"]["loss"]["params"]["label_as_categories"],
			from_logits=False);

	elif(name == "crossentropy"):
		system_dict["local"]["criterion"] = mx.gluon.loss.SoftmaxCrossEntropyLoss(
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"],
			axis=system_dict["hyper-parameters"]["loss"]["params"]["axis_to_sum_over"],
			sparse_label=system_dict["hyper-parameters"]["loss"]["params"]["label_as_categories"],
			from_logits=True);

	elif(name == "sigmoidbinarycrossentropy"):
		system_dict["local"]["criterion"] = mx.gluon.loss.SigmoidBinaryCrossEntropyLoss(
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"],
			from_sigmoid=False);

	elif(name == "binarycrossentropy"):
		system_dict["local"]["criterion"] = mx.gluon.loss.SigmoidBinaryCrossEntropyLoss(
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"],
			from_sigmoid=True);

	elif(name == "kldiv"):
		system_dict["local"]["criterion"] = mx.gluon.loss.KLDivLoss(
			from_logits=system_dict["hyper-parameters"]["loss"]["params"]["log_pre_applied"], 
			axis=system_dict["hyper-parameters"]["loss"]["params"]["axis_to_sum_over"], 
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"], 
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"]);

	elif(name == "poissonnll"):
		system_dict["local"]["criterion"] = mx.gluon.loss.PoissonNLLLoss(
			from_logits=system_dict["hyper-parameters"]["loss"]["params"]["log_pre_applied"],
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"]);

	elif(name == "huber"):
		system_dict["local"]["criterion"] = mx.gluon.loss.HuberLoss(
			rho=system_dict["hyper-parameters"]["loss"]["params"]["threshold_for_mean_estimator"],
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"]);

	elif(name == "hinge"):
		system_dict["local"]["criterion"] = mx.gluon.loss.HingeLoss(
			margin=system_dict["hyper-parameters"]["loss"]["params"]["margin"],
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"]);	

	elif(name == "squaredhinge"):
		system_dict["local"]["criterion"] = mx.gluon.loss.SquaredHingeLoss(
			margin=system_dict["hyper-parameters"]["loss"]["params"]["margin"],
			weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"],
			batch_axis=system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"]);

	return system_dict;


from monk.system.graphs.imports import *
from monk.system.imports import *



@accepts([list, type(np.array([1, 2]))], [list, type(np.array([1, 2]))], str, show_img=bool, save_img=bool, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def create_plot(labels, value, log_dir, show_img=False, save_img=False):
    '''
    Create bar plot

    Args:
        labels (list): X - Axis labels
        value (list): Coresponding Y-Axis values
        log_dir (str): Path to folder to save plot 
        show_img (bool): If True, plot is displayed
        save_img (bool): If True, plot is saved.
        verbose (str): If True, prints logs for analysis 

    Returns:
        None
    '''
    objects = tuple(labels)
    y_pos = np.arange(len(objects))
    plt.bar(y_pos, value, align='center', alpha=0.5, width=0.5)
    plt.xticks(y_pos, objects, rotation='vertical', linespacing=0.9, fontsize='xx-small')
    plt.ylabel('Time - Seconds')
    plt.autoscale()
    plt.title("Training time plot");
    plt.tight_layout()

    if(save_img):
        file_name = log_dir + "/classes_per_image.png" 
        plt.savefig(file_name)
    if(show_img):
        plt.show()    
    plt.clf()



@accepts([list, type(np.array([1, 2]))], dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=True)
def training_time_plot(data, system_dict):
    '''
    Create training-time bar plot

    Args:
        data (list): Entire data
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    labels = [ '\n'.join(wrap(l, 20)) for l in system_dict["local"]["project_experiment_list"] ]
    objects = tuple(labels)
    y_pos = np.arange(len(objects))
    value = []
    for i in range(len(data)):
        if(data[i]["training"]["status"]):
            minutes, seconds = data[i]["training"]["outputs"]["training_time"].split(" ");
            minutes = int(minutes[:len(minutes)-1]);
            seconds = int(seconds[:len(seconds)-1]);
            value.append(minutes*60 + seconds);
        else:
            value.append(0);
    plt.bar(y_pos, value, align='center', alpha=0.5)
    plt.xticks(y_pos, objects, rotation='vertical')
    plt.ylabel('Time - Seconds')
    plt.autoscale()
    plt.title("Training time plot");
    plt.tight_layout()
    file_name = system_dict["comparison_dir"] + "stats_training_time.png" 
    plt.savefig(file_name)
    plt.clf()


@accepts([list, type(np.array([1, 2]))], dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=True)
def max_accuracy_plot(data, system_dict):
    '''
    Create max-accuracy bar plot

    Args:
        data (list): Entire data
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    labels = [ '\n'.join(wrap(l, 20)) for l in system_dict["local"]["project_experiment_list"] ]
    objects = tuple(labels)
    y_pos = np.arange(len(objects))
    value = []
    for i in range(len(data)):
        if(data[i]["training"]["status"]):
            acc = float(data[i]["training"]["outputs"]["best_val_acc"])*100;
            value.append(acc);
        else:
            print("Testing for {} - Incomplete".format(system_dict["local"]["project_experiment_list"][i]));
            value.append(0);
    plt.bar(y_pos, value, align='center', alpha=0.5)
    plt.xticks(y_pos, objects, rotation='vertical')
    plt.ylabel('Accuracy (%)')
    plt.autoscale()
    plt.title("Best validation accuracy plot");
    plt.tight_layout()
    file_name = system_dict["comparison_dir"] + "stats_best_val_acc.png" 
    plt.savefig(file_name)
    plt.clf()


@accepts([list, type(np.array([1, 2]))], dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=True)
def max_gpu_usage_plot(data, system_dict):
    '''
    Create gpu-usage bar plot

    Args:
        data (list): Entire data
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    labels = [ '\n'.join(wrap(l, 20)) for l in system_dict["local"]["project_experiment_list"] ]
    objects = tuple(labels)
    y_pos = np.arange(len(objects))
    value = []
    for i in range(len(data)):
        if(data[i]["training"]["status"]):
            acc = int(data[i]["training"]["outputs"]["max_gpu_usage"].split(" ")[0]);
            value.append(acc);
        else:
            value.append(0);
    plt.bar(y_pos, value, align='center', alpha=0.5)
    plt.xticks(y_pos, objects, rotation='vertical')
    plt.ylabel('Usage (Mb)')
    plt.autoscale()
    plt.title("Max GPU Memory Usage plot");
    plt.tight_layout()
    file_name = system_dict["comparison_dir"] + "stats_max_gpu_usage.png" 
    plt.savefig(file_name)
    plt.clf()


from monk.system.graphs.imports import *
from monk.system.imports import * 


@accepts([list, type(np.array([1, 2]))], [list, type(np.array([1, 2]))], str, show_img=bool, save_img=bool, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def create_train_test_plots_accuracy(plots, labels, log_dir, show_img=False, save_img=False):
    '''
    Create line graphs for training and testing accuracies

    Args:
        plots (list): plots[0] contains training data
                        plots[1] contains validation data
        labels (list): Coresponding X-Axis labels
        log_dir (str): Path to folder to save plot 
        show_img (bool): If True, plot is displayed
        save_img (bool): If True, plot is saved.
        verbose (str): If True, prints logs for analysis 

    Returns:
        None
    '''
    plt.plot(plots[0], marker='o', label='Training')
    plt.plot(plots[1], marker='x', label='Validation')
    plt.gca().legend(('Training','Validation'))
    plt.xlabel(labels[0]);
    plt.ylabel(labels[1]);
    file_name = log_dir + "train_val_accuracy.png" 
    plt.savefig(file_name)
    plt.clf()


@accepts([list, type(np.array([1, 2]))], [list, type(np.array([1, 2]))], str, show_img=bool, save_img=bool, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def create_train_test_plots_loss(plots, labels, log_dir, show_img=False, save_img=False):
    '''
    Create line graphs for training and testing losses

    Args:
        plots (list): plots[0] contains training data
                        plots[1] contains validation data
        labels (list): Coresponding X-Axis labels
        log_dir (str): Path to folder to save plot 
        show_img (bool): If True, plot is displayed
        save_img (bool): If True, plot is saved.
        verbose (str): If True, prints logs for analysis 

    Returns:
        None
    '''
    plt.plot(plots[0], marker='o', label='Training')
    plt.plot(plots[1], marker='x', label='Validation')
    plt.gca().legend(('Training','Validation'))
    plt.xlabel(labels[0]);
    plt.ylabel(labels[1]);
    file_name = log_dir + "train_val_loss.png" 
    plt.savefig(file_name)
    plt.clf()


@accepts([list, type(np.array([1, 2]))], dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=True)
def training_accuracy_curve(data, system_dict):
    '''
    Create training accuracy line graph

    Args:
        data (list): Entire data
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    plt.figure(figsize=(20,10))
    for i in range(len(data)):
        if(not data[i]["training"]["status"]):
            print("Training for {} - Incomplete".format(self.project_experiment_list[i]));
            value = np.zeros(1);
            xkcd = mcd.XKCD_COLORS["xkcd:" + overlap[i]].upper()
            plt.plot(value, marker='o', label=system_dict["local"]["project_experiment_list"][i], color=xkcd, linewidth=5.0);
        else:
            value = np.load(data[i]["training"]["outputs"]["log_train_acc_history_relative"], allow_pickle=True);
            xkcd = mcd.XKCD_COLORS["xkcd:" + overlap[i]].upper()
            plt.plot(value, marker='o', label=system_dict["local"]["project_experiment_list"][i], color=xkcd, linewidth=5.0);
    plt.gca().legend(tuple(system_dict["local"]["project_experiment_list"]));
    plt.xlabel("Epoch Num");
    plt.ylabel("Accuracy");
    plt.title("Training Accuracy Curve");
    plt.autoscale()
    file_name = system_dict["comparison_dir"] + "train_accuracy.png" 
    plt.savefig(file_name)
    plt.clf()


@accepts([list, type(np.array([1, 2]))], dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=True)
def validation_accuracy_curve(data, system_dict):
    '''
    Create validation accuracy line graph

    Args:
        data (list): Entire data
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    plt.figure(figsize=(20,10))
    for i in range(len(data)):
        if(not data[i]["training"]["status"]):
            value = np.zeros(1);
            xkcd = mcd.XKCD_COLORS["xkcd:" + overlap[i]].upper()
            plt.plot(value, marker='o', label=system_dict["local"]["project_experiment_list"][i], color=xkcd, linewidth=5.0);
        else:
            value = np.load(data[i]["training"]["outputs"]["log_val_acc_history_relative"], allow_pickle=True);
            xkcd = mcd.XKCD_COLORS["xkcd:" + overlap[i]].upper()
            plt.plot(value, marker='x', label=system_dict["local"]["project_experiment_list"][i], color=xkcd, linewidth=5.0)
    plt.gca().legend(tuple(system_dict["local"]["project_experiment_list"]));
    plt.xlabel("Epoch Num");
    plt.ylabel("Accuracy");
    plt.title("Validation Accuracy Curve");
    plt.autoscale()
    file_name = system_dict["comparison_dir"] + "val_accuracy.png" 
    plt.savefig(file_name)
    plt.clf()


@accepts([list, type(np.array([1, 2]))], dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=True)
def training_loss_curve(data, system_dict):
    '''
    Create training loss line graph

    Args:
        data (list): Entire data
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    plt.figure(figsize=(20,10))
    for i in range(len(data)):
        if(not data[i]["training"]["status"]):
            value = np.zeros(1);
            xkcd = mcd.XKCD_COLORS["xkcd:" + overlap[i]].upper()
            plt.plot(value, marker='o', label=system_dict["local"]["project_experiment_list"][i], color=xkcd, linewidth=5.0);
        else:
            value = np.load(data[i]["training"]["outputs"]["log_train_loss_history_relative"], allow_pickle=True);
            xkcd = mcd.XKCD_COLORS["xkcd:" + overlap[i]].upper()
            plt.plot(value, marker='o', label=system_dict["local"]["project_experiment_list"][i], color=xkcd, linewidth=5.0)
    plt.gca().legend(tuple(system_dict["local"]["project_experiment_list"]));
    plt.xlabel("Epoch Num");
    plt.ylabel("Loss");
    plt.title("Training Loss Curve");
    plt.autoscale()
    file_name = system_dict["comparison_dir"] + "train_loss.png" 
    plt.savefig(file_name)
    plt.clf()


@accepts([list, type(np.array([1, 2]))], dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=True)
def validation_loss_curve(data, system_dict):
    '''
    Create validation loss line graph

    Args:
        data (list): Entire data
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    plt.figure(figsize=(20,10))
    for i in range(len(data)):
        if(not data[i]["training"]["status"]):
            value = np.zeros(1);
            xkcd = mcd.XKCD_COLORS["xkcd:" + overlap[i]].upper()
            plt.plot(value, marker='o', label=system_dict["local"]["project_experiment_list"][i], color=xkcd, linewidth=5.0);
        else:
            value = np.load(data[i]["training"]["outputs"]["log_val_loss_history_relative"], allow_pickle=True);
            xkcd = mcd.XKCD_COLORS["xkcd:" + overlap[i]].upper()
            plt.plot(value, marker='o', label=system_dict["local"]["project_experiment_list"][i], color=xkcd, linewidth=5.0)
    plt.gca().legend(tuple(system_dict["local"]["project_experiment_list"]));
    plt.xlabel("Epoch Num");
    plt.ylabel("Loss");
    plt.title("Validation Loss Curve");
    plt.autoscale()
    file_name = system_dict["comparison_dir"] + "val_loss.png" 
    plt.savefig(file_name)
    plt.clf()







import os
import sys

import psutil

from monk.gluon_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_unit_tests.gluon.common import print_start
from monk.pip_unit_tests.gluon.common import print_status

import mxnet as mx
import numpy as np
from monk.gluon.losses.return_loss import load_loss


def test_loss_huber(system_dict):
    forward = True;

    test = "test_loss_huber";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf = prototype(verbose=0);
            gtf.Prototype("sample-project-1", "sample-experiment-1");

            label = np.random.rand(1, 5);
            label = mx.nd.array(label);

            y = np.random.rand(1, 5);
            y = mx.nd.array(y);

            gtf.loss_huber();
            load_loss(gtf.system_dict);
            loss_obj = gtf.system_dict["local"]["criterion"];
            loss_val = loss_obj(y, label);           

            system_dict["successful_tests"] += 1;
            print_status("Pass");

        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    return system_dict



from monk.gluon.models.imports import *
from monk.system.imports import *
from monk.gluon.models.layers import get_layer


@accepts("self", bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_parameter_requires_grad(finetune_net, freeze_base_network):
    '''
    Freeze based network as per params set

    Args:
        finetune_net (network): Model network
        freeze_base_network (bool): If True, all trainable params are freezed

    Returns:
        network: Updated Model network
    '''
    if(freeze_base_network):
        for param in finetune_net.collect_params().values():
            param.grad_req = 'null';
    return finetune_net


@accepts(dict, activation=str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def get_final_layer(network_layer, activation='relu'):
    '''
    Setup final layer

    Args:
        network_layer (dict): Dictionary containing all parameters for layer
        activation (str): Activation function to use  
                            - Available functions: 'relu', 'sigmoid', 'tanh', 'softrelu', 'softsign'

    Returns:
        layer: Designed layer
    '''
    act_list = ['relu', 'sigmoid', 'tanh', 'softrelu', 'softsign'];

    if(activation not in act_list):
        print("Final activation must be in set: {}".format(act_list));
        print("");
    else:
        layer = nn.Dense(network_layer["params"]["out_features"], weight_initializer=init.Xavier(), activation=activation);
        return layer;


@accepts("self", list, int, set=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def create_final_layer(finetune_net, custom_network, num_classes, set=1):
    '''
    Create final sub-network 

    Args:
        finetune_net (network): Initial base network
        custom_network (list):  List of dicts containing details on appeded layers to base netwoek in transfer learning
        num_classes (int): Number of classes in the dataset
        set (int): Select the right set to find the details of outermost layer

    Returns:
        network: Updated base network with appended custom additions
    '''
    last_layer_name = custom_network[len(custom_network)-1]["name"];

    if(set==1):
        if(last_layer_name == "linear"):
            with finetune_net.name_scope():
                for i in range(len(custom_network)-1):
                    layer = get_layer(custom_network[i]);
                    finetune_net.features.add(layer);
                    finetune_net.features[len(finetune_net.features)-1].initialize(init.Xavier(), ctx = ctx);
                finetune_net.output = get_layer(custom_network[len(custom_network)-1])
                finetune_net.output.initialize(init.Xavier(), ctx = ctx)
        else:
            with finetune_net.name_scope():
                for i in range(len(custom_network)-2):
                    layer = get_layer(custom_network[i]);
                    finetune_net.features.add(layer);
                    finetune_net.features[len(finetune_net.features)-1].initialize(init.Xavier(), ctx = ctx);
                finetune_net.output = get_final_layer(custom_network[len(custom_network)-2], activation=custom_network[len(custom_network)-1]['name']);
                finetune_net.output.initialize(init.Xavier(), ctx = ctx)


    if(set==2):
        net = nn.HybridSequential();
        with net.name_scope():
            for i in range(len(custom_network)):
                layer = get_layer(custom_network[i]);
                net.add(layer);
        with finetune_net.name_scope():
            finetune_net.output = net; 
            finetune_net.output.initialize(init.Xavier(), ctx = ctx) 


    if(set==3):
        msg = "Custom model addition for - Set 3 models: Not Implemented.\n";
        msg += "Set 3 models - {}\n".format(set3);
        msg += "Ignoring added layers\n";
        ConstraintWarning(msg);

    return finetune_net;



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def model_to_device(system_dict):
    '''
    Load model weights on device - cpu or gpu 

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    if(not mx.context.num_gpus()):
        system_dict["local"]["ctx"] = [mx.cpu()];
    else:
        if(system_dict["model"]["params"]["use_gpu"]):
            system_dict["local"]["ctx"] = [mx.gpu(0)];
        else:
            system_dict["local"]["ctx"] = [mx.cpu()];

    system_dict["local"]["model"].collect_params().reset_ctx(system_dict["local"]["ctx"])
    system_dict["local"]["model"].hybridize()

    return system_dict;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def print_grad_stats(system_dict):
    '''
    Print details on which layers are trainable

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        None 
    '''
    print("Model - Gradient Statistics");
    i = 1;
    for param in system_dict["local"]["model"].collect_params().values():
        if(i%2 != 0):
            print("    {}. {} Trainable - {}".format(i//2+1, param, param.grad_req ));
        i += 1;
    print("");


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def get_num_layers(system_dict):
    '''
    Get number of potentially trainable layers

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    if(system_dict["model"]["type"] == "pretrained"):
        num_layers = 0;
        complete_list = [];
        for names in system_dict["local"]["model"].collect_params().keys():
            layer_name = '_'.join(names.split("_")[:-1]);
            if("running" in layer_name):
                layer_name = '_'.join(layer_name.split("_")[:-1]);
            if(layer_name not in complete_list):
                complete_list.append(layer_name);
        system_dict["model"]["params"]["num_layers"] = len(np.unique(complete_list));
        return system_dict;
    else:
        num_layers = 0;
        complete_list = [];
        for names in system_dict["local"]["model"].collect_params().keys():
            complete_list.append(names.split("_")[0]);
        system_dict["model"]["params"]["num_layers"] = len(np.unique(complete_list));
        return system_dict;



@accepts(int, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def freeze_layers(num, system_dict):
    '''
    Main function responsible to freeze layers in network

    Args:
        num (int): Number of layers to freeze
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    if(system_dict["model"]["type"] == "pretrained"):
        num_freeze = num;
        num_freezed = 0;
        training_list = system_dict["local"]["params_to_update"];

        grad_req= "null";
        current_name = training_list[0];
        complete_list = [];
        for param in system_dict["local"]["model"].collect_params().values():
            layer_name = '_'.join(param.name.split("_")[:-1]);
            if("running" in layer_name):
                layer_name = '_'.join(layer_name.split("_")[:-1]);
            if(layer_name not in complete_list):
                num_freezed += 1;
                if(num_freezed == num_freeze):
                    grad_req = "write";
                param.grad_req = grad_req;
                complete_list.append(layer_name)
            else:
                param.grad_req = grad_req;

        current_name = "";
        ip = 0;
        system_dict["local"]["params_to_update"] = [];
        complete_list = [];
        for param in system_dict["local"]["model"].collect_params().values():
            if(ip==0):
                current_name = param.name.split("_")[:-1];
                if("running" in current_name):
                    current_name = '_'.join(current_name[:-1]);
                if(param.grad_req == "write"):
                    if(current_name not in complete_list):
                        complete_list.append(current_name);
                    system_dict["local"]["params_to_update"].append(current_name);
            else:
                if(current_name != param.name.split("_")[:-1]):
                    current_name = param.name.split("_")[:-1];
                    if("running" in current_name):
                        current_name = '_'.join(current_name[:-1]);
                    if(param.grad_req == "write"):
                        if(current_name not in complete_list):
                            complete_list.append(current_name);
                        system_dict["local"]["params_to_update"].append(current_name);
            ip += 1;
        system_dict["model"]["params"]["num_params_to_update"] = len(complete_list);
        system_dict["model"]["status"] = True;

        return system_dict;

    else:
        num_freeze = num;
        num_freezed = 0;
        training_list = system_dict["local"]["params_to_update"];


        current_name = training_list[0];
        for param in system_dict["local"]["model"].collect_params().values():
            if(current_name != param.name.split("_")[:-1]):
                num_freezed += 1;
                if(num_freezed == num_freeze):
                    break;
                param.grad_req = "null";
                current_name = param.name.split("_")[:-1];
            else:
                param.grad_req = "null";


        current_name = "";
        ip = 0;
        system_dict["local"]["params_to_update"] = [];
        complete_list = [];
        for param in system_dict["local"]["model"].collect_params().values():
            if(ip==0):
                current_name = '_'.join(param.name.split("_")[:-1]);
                if("running" in current_name):
                    current_name = '_'.join(current_name[:-1]);
                if(param.grad_req == "write"):
                    if(current_name not in complete_list):
                        complete_list.append(current_name);
                    system_dict["local"]["params_to_update"].append(current_name);
            else:
                if(current_name != '_'.join(param.name.split("_")[:-1])):
                    current_name = '_'.join(param.name.split("_")[:-1]);
                    if("running" in current_name):
                        current_name = '_'.join(current_name[:-1]);
                    if(param.grad_req == "write"):
                        if(current_name not in complete_list):
                            complete_list.append(current_name);
                        system_dict["local"]["params_to_update"].append(current_name);
            ip += 1;
        system_dict["model"]["params"]["num_params_to_update"] = len(system_dict["local"]["params_to_update"]);
        system_dict["model"]["status"] = True;



        return system_dict;


@accepts(dict, list, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def get_layer_uid(network_stack, count):
    '''
    Get a unique name for layer in custom network development

    Args:
        network_stack (list): List of list containing custom network details
        count (dict): a unique dictionary mapping number of every type of layer in the network
        system_dict (dict): System dict containing system state and parameters

    Returns:
        str: layer unique name
        dict: updated layer type mapper count
    '''
    if network_stack["uid"]:
        return network_stack["uid"], count;
    else:
        index = layer_names.index(network_stack["name"]);
        network_name = names[index] + str(count[index]);
        count[index] += 1;
        return network_name, count;



from monk.tf_keras_1.models.imports import *
from monk.system.imports import *
from monk.tf_keras_1.models.initializers import get_initializer



@accepts(dict, "self", post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def get_layer(network_layer, inp):
    '''
    Get layer for appending it to transfer learning base model

    Args:
        network_layer (dict): Dictionary conatining all params relating to the layer
        inp (keras placeholder): Placeholder input for the layer

    Returns:
        Keras placeholder: Placeholder output from the layer
    '''
    layer_name = network_layer["name"];
    layer_params = network_layer["params"];


    if(layer_name == "linear"):
        out = krl.Dense(layer_params["out_features"])(inp)
    
    elif(layer_name == "dropout"):
        out = krl.Dropout(rate=1-layer_params["p"])(inp)

    elif(layer_name == "globalaveragepooling"):
        out = krl.GlobalAveragePooling2D()(inp);

    elif(layer_name == "flatten"):
        out = krl.Flatten()(inp);
    
    elif(layer_name == "elu"):
        out = krl.ELU(alpha=layer_params["alpha"])(inp);
    
    elif(layer_name == "leakyrelu"):
        out = krl.LeakyReLU(alpha=layer_params["alpha"])(inp);
    
    elif(layer_name == "prelu"):
        init = kri.Constant(value=layer_params["init"])
        out = krl.PReLU(alpha_initializer=init)(inp);
    
    elif(layer_name == "relu"):
        out = krl.ReLU()(inp);

    elif(layer_name == "selu"):
        out = krl.selu(inp);

    elif(layer_name == "sigmoid"):
        out = krl.sigmoid(inp);

    elif(layer_name == "softplus"):
        out = krl.softplus(inp);

    elif(layer_name == "softsign"):
        out = krl.softsign(inp);

    elif(layer_name == "tanh"):
        out = krl.tanh(inp);

    elif(layer_name == "threshold"):
        out = krl.ThresholdedReLU(theta=layer_params["threshold"])(inp);

    elif(layer_name == "softmax"):
        out = krl.Softmax()(inp);


    return out;



@accepts(dict, num_neurons=int, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def layer_linear(system_dict, num_neurons=512, final_layer=False):
    '''
    Append dense (fully connected) layer to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        num_neurons (int): Number of neurons in the dense layer
        final_layer (bool): If True, then number of neurons are directly set as number of classes in dataset for single label type classification

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "linear";
    tmp["params"] = {};
    tmp["params"]["out_features"] = num_neurons;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, probability=float, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def layer_dropout(system_dict, probability=0.5, final_layer=False):
    '''
    Append dropout layer to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        probability (float): Droping probability of neurons in next layer
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "dropout";
    tmp["params"] = {};
    tmp["params"]["p"] = probability;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def layer_globalaveragepooling(system_dict, final_layer=False):
    '''
    Append global average pooling layer to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "globalaveragepooling";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def layer_flatten(system_dict, final_layer=False):
    '''
    Append flatten layer to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "flatten";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, negative_slope=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_leakyrelu(system_dict, negative_slope=0.01, final_layer=False):
    '''
    Append Leaky - ReLU activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        negative_slope (float): Multiplicatve factor towards negative spectrum of real numbers.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "leakyrelu";
    tmp["params"] = {};
    tmp["params"]["negative_slope"] = negative_slope; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, num_parameters=int, init=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_prelu(system_dict, num_parameters=1, init=0.25, final_layer=False):
    '''
    Append Learnable parameterized rectified linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        init (float): Initialization value for multiplicatve factor towards negative spectrum of real numbers.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "prelu";
    tmp["params"] = {};
    tmp["params"]["num_parameters"] = num_parameters; 
    tmp["params"]["init"] = init; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, alpha=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_elu(system_dict, alpha=1.0, final_layer=False):
    '''
    Append exponential linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        alpha (float): Multiplicatve factor.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "elu";
    tmp["params"] = {};
    tmp["params"]["alpha"] = alpha;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, [int, float], [int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_threshold(system_dict, threshold, value, final_layer=False):
    '''
    Append threshold activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        threshold (float): thereshold limit value
        value (float): replacement value if input is off-limits
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "threshold";
    tmp["params"] = {};
    tmp["params"]["value"] = value; 
    tmp["params"]["threshold"] = threshold;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;





@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softmax(system_dict, final_layer=False):
    '''
    Append softmax activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softmax";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_relu(system_dict, final_layer=False):
    '''
    Append rectified linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "relu";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_selu(system_dict, final_layer=False):
    '''
    Append scaled exponential linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "selu";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, beta=[int, float], threshold=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softplus(system_dict, beta=1, threshold=20, final_layer=False):
    '''
    Append softplus activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        beta (int): Multiplicative factor
        threshold (int): softplus (thresholded relu) limit 
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softplus";
    tmp["params"] = {};
    tmp["params"]["beta"] = beta; 
    tmp["params"]["threshold"] = threshold; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softsign(system_dict, final_layer=False):
    '''
    Append softsign activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softsign";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_tanh(system_dict, final_layer=False):
    '''
    Append tanh activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "tanh";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_sigmoid(system_dict, final_layer=False):
    '''
    Append sigmoid activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "sigmoid";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;









@accepts(dict, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def custom_model_get_layer(network_layer, network_initializer):
    '''
    Get layer for custom network design

    Args:
        network_layer (dict): Network layer as dict
        network_initializer (initialization param): Initializer for the layer

    Returns:
        neural network layer: Actual layer in native backend library
    '''
    layer_name = network_layer["name"];
    layer_params = network_layer["params"];
    if(layer_name == "convolution1d"):
        return custom_model_layer_convolution1d(layer_params, network_initializer);
    elif(layer_name == "convolution2d"):
        return custom_model_layer_convolution2d(layer_params, network_initializer);
    elif(layer_name == "convolution3d"):
        return custom_model_layer_convolution3d(layer_params, network_initializer);

    elif(layer_name == "transposed_convolution2d"):
        return custom_model_layer_transposed_convolution2d(layer_params, network_initializer);
    elif(layer_name == "transposed_convolution3d"):
        return custom_model_layer_transposed_convolution3d(layer_params, network_initializer);

    elif(layer_name == "max_pooling1d"):
        return custom_model_layer_max_pooling1d(layer_params);
    elif(layer_name == "max_pooling2d"):
        return custom_model_layer_max_pooling2d(layer_params);
    elif(layer_name == "max_pooling3d"):
        return custom_model_layer_max_pooling3d(layer_params);

    elif(layer_name == "average_pooling1d"):
        return custom_model_layer_average_pooling1d(layer_params);
    elif(layer_name == "average_pooling2d"):
        return custom_model_layer_average_pooling2d(layer_params);
    elif(layer_name == "average_pooling3d"):
        return custom_model_layer_average_pooling3d(layer_params);

    elif(layer_name == "global_max_pooling1d"):
        return custom_model_layer_global_max_pooling1d(layer_params);
    elif(layer_name == "global_max_pooling2d"):
        return custom_model_layer_global_max_pooling2d(layer_params);
    elif(layer_name == "global_max_pooling3d"):
        return custom_model_layer_global_max_pooling3d(layer_params);

    elif(layer_name == "global_average_pooling1d"):
        return custom_model_layer_global_average_pooling1d(layer_params);
    elif(layer_name == "global_average_pooling2d"):
        return custom_model_layer_global_average_pooling2d(layer_params);
    elif(layer_name == "global_average_pooling3d"):
        return custom_model_layer_global_average_pooling3d(layer_params);

    elif(layer_name == "flatten"):
        return custom_model_layer_flatten(layer_params);
    elif(layer_name == "fully_connected"):
        return custom_model_layer_fully_connected(layer_params, network_initializer);
    elif(layer_name == "dropout"):
        return custom_model_layer_dropout(layer_params);
    elif(layer_name == "identity"):
        return custom_model_layer_identity(layer_params);

    elif(layer_name == "batch_normalization"):
        return custom_model_layer_batch_normalization(layer_params);

    elif(layer_name == "relu"):
        return custom_model_activation_relu(layer_params);
    elif(layer_name == "softmax"):
        return custom_model_activation_softmax(layer_params);
    elif(layer_name == "thresholded_relu"):
        return custom_model_activation_thresholded_relu(layer_params);
    elif(layer_name == "elu"):
        return custom_model_activation_elu(layer_params);
    elif(layer_name == "prelu"):
        return custom_model_activation_prelu(layer_params);
    elif(layer_name == "leaky_relu"):
        return custom_model_activation_leaky_relu(layer_params);
    elif(layer_name == "selu"):
        return custom_model_activation_selu(layer_params);
    elif(layer_name == "softplus"):
        return custom_model_activation_softplus(layer_params);
    elif(layer_name == "softsign"):
        return custom_model_activation_softsign(layer_params);
    elif(layer_name == "tanh"):
        return custom_model_activation_tanh(layer_params);
    elif(layer_name == "sigmoid"):
        return custom_model_activation_sigmoid(layer_params);
    elif(layer_name == "hard_sigmoid"):
        return custom_model_activation_hard_sigmoid(layer_params);






@accepts(dict, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution1d(params, network_initializer):
    '''
    Append 1d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-convolution layer
    '''
    out_channels = params["output_channels"];
    kernel_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "same"; #causal

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    dilation_rate=params["dilation"];
    activation=None;
    use_bias = params["use_bias"];
    kernel_initializer=get_initializer(network_initializer)
    bias_initializer='zeros';
    kernel_regularizer=None;
    bias_regularizer=None;
    activity_regularizer=None;
    kernel_constraint=None;
    bias_constraint=None;


    layer = keras.layers.Conv1D(out_channels, kernel_size, strides=strides, 
                            padding=padding, data_format=data_format, 
                            dilation_rate=dilation_rate, activation=activation, use_bias=use_bias, 
                            kernel_initializer=kernel_initializer, bias_initializer=bias_initializer, 
                            kernel_regularizer=kernel_regularizer, bias_regularizer=bias_regularizer, 
                            activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint, 
                            bias_constraint=bias_constraint)

    return layer


@accepts(dict, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution2d(params, network_initializer):
    '''
    Append 2d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-convolution layer
    '''
    out_channels = params["output_channels"];
    kernel_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal"; #causal

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    dilation_rate=params["dilation"];
    activation=None;
    use_bias = params["use_bias"];
    kernel_initializer=get_initializer(network_initializer);
    bias_initializer='zeros';
    kernel_regularizer=None;
    bias_regularizer=None;
    activity_regularizer=None;
    kernel_constraint=None;
    bias_constraint=None;


    layer = keras.layers.Conv2D(out_channels, kernel_size, strides=strides, 
                            padding=padding, data_format=data_format, 
                            dilation_rate=dilation_rate, activation=activation, use_bias=use_bias, 
                            kernel_initializer=kernel_initializer, bias_initializer=bias_initializer, 
                            kernel_regularizer=kernel_regularizer, bias_regularizer=bias_regularizer, 
                            activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint, 
                            bias_constraint=bias_constraint)

    return layer



@accepts(dict, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution3d(params, network_initializer):
    '''
    Append 3d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-convolution layer
    '''
    out_channels = params["output_channels"];
    kernel_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "same"; #causal

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    dilation_rate=params["dilation"];
    activation=None;
    use_bias = params["use_bias"];
    kernel_initializer=get_initializer(network_initializer);
    bias_initializer='zeros';
    kernel_regularizer=None;
    bias_regularizer=None;
    activity_regularizer=None;
    kernel_constraint=None;
    bias_constraint=None;


    layer = keras.layers.Conv3D(out_channels, kernel_size, strides=strides, 
                            padding=padding, data_format=data_format, 
                            dilation_rate=dilation_rate, activation=activation, use_bias=use_bias, 
                            kernel_initializer=kernel_initializer, bias_initializer=bias_initializer, 
                            kernel_regularizer=kernel_regularizer, bias_regularizer=bias_regularizer, 
                            activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint, 
                            bias_constraint=bias_constraint)

    return layer



@accepts(dict, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_transposed_convolution2d(params, network_initializer):
    '''
    Append 2d-transposed-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-transposed-convolution layer
    '''
    out_channels = params["output_channels"];
    kernel_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal";

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    dilation_rate=params["dilation"];
    activation=None;
    use_bias = params["use_bias"];
    kernel_initializer=get_initializer(network_initializer);
    bias_initializer='zeros';
    kernel_regularizer=None;
    bias_regularizer=None;
    activity_regularizer=None;
    kernel_constraint=None;
    bias_constraint=None;


    layer = keras.layers.Conv2DTranspose(out_channels, kernel_size, strides=strides, 
                            padding=padding, data_format=data_format, 
                            dilation_rate=dilation_rate, activation=activation, use_bias=use_bias, 
                            kernel_initializer=kernel_initializer, bias_initializer=bias_initializer, 
                            kernel_regularizer=kernel_regularizer, bias_regularizer=bias_regularizer, 
                            activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint, 
                            bias_constraint=bias_constraint)

    return layer


@accepts(dict, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_transposed_convolution3d(params, network_initializer):
    '''
    Append 3d-transposed-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-transposed-convolution layer
    '''
    out_channels = params["output_channels"];
    kernel_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal";

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    dilation_rate=params["dilation"];
    activation=None;
    use_bias = params["use_bias"];
    kernel_initializer=get_initializer(network_initializer);
    bias_initializer='zeros';
    kernel_regularizer=None;
    bias_regularizer=None;
    activity_regularizer=None;
    kernel_constraint=None;
    bias_constraint=None;


    layer = keras.layers.Conv3DTranspose(out_channels, kernel_size, strides=strides, 
                            padding=padding, data_format=data_format, 
                            dilation_rate=dilation_rate, activation=activation, use_bias=use_bias, 
                            kernel_initializer=kernel_initializer, bias_initializer=bias_initializer, 
                            kernel_regularizer=kernel_regularizer, bias_regularizer=bias_regularizer, 
                            activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint, 
                            bias_constraint=bias_constraint)
    
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling1d(params):
    '''
    Append 1d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-max-pooling layer
    '''
    pool_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal";

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';


    layer = keras.layers.MaxPooling1D(pool_size=pool_size, strides=strides, 
                                    padding=padding, data_format=data_format)

    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling2d(params):
    '''
    Append 2d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-max-pooling layer
    '''

    pool_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal";

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';


    layer = keras.layers.MaxPooling2D(pool_size=pool_size, strides=strides, 
                                    padding=padding, data_format=data_format)

    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling3d(params):
    '''
    Append 3d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-max-pooling layer
    '''
    pool_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal";

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';


    layer = keras.layers.MaxPooling3D(pool_size=pool_size, strides=strides, 
                                    padding=padding, data_format=data_format)

    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling1d(params):
    '''
    Append 1d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-average-pooling layer
    '''

    pool_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal";

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';


    layer = keras.layers.AveragePooling1D(pool_size=pool_size, strides=strides, 
                                    padding=padding, data_format=data_format)

    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling2d(params):
    '''
    Append 2d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-average-pooling layer
    '''

    pool_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal";

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';


    layer = keras.layers.AveragePooling2D(pool_size=pool_size, strides=strides, 
                                    padding=padding, data_format=data_format)

    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling3d(params):
    '''
    Append 3d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-average-pooling layer
    '''

    pool_size=params["kernel_size"];
    strides=params["stride"];
    if(params["padding"] == "in_eq_out"):
        padding = "same";
    elif(params["padding"] == 0):
        padding = "valid";
    else:
        padding = "causal";

    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';


    layer = keras.layers.AveragePooling3D(pool_size=pool_size, strides=strides, 
                                    padding=padding, data_format=data_format)

    return layer



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling1d(params):
    '''
    Append 1d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-global-max-pooling layer
    '''
    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    layer = keras.layers.GlobalMaxPooling1D(data_format=data_format);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling2d(params):
    '''
    Append 2d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-global-max-pooling layer
    '''
    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';
    
    layer = keras.layers.GlobalMaxPooling2D(data_format=data_format);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling3d(params):
    '''
    Append 3d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-global-max-pooling layer
    '''
    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';
    
    layer = keras.layers.GlobalMaxPooling3D(data_format=data_format);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling1d(params):
    '''
    Append 1d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-global-average-pooling layer
    '''
    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    layer = keras.layers.GlobalAveragePooling1D(data_format=data_format);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling2d(params):
    '''
    Append 2d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-global-average-pooling layer
    '''
    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    layer = keras.layers.GlobalAveragePooling2D(data_format=data_format);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling3d(params):
    '''
    Append 3d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-global-average-pooling layer
    '''
    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    layer = keras.layers.GlobalAveragePooling3D(data_format=data_format);
    return layer


@accepts(dict, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_flatten(params):
    '''
    Append flatten to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: flatten layer
    '''
    if(params["layout"][-1] == "C"):
        data_format='channels_last';
    else:
        data_format='channels_first';

    layer = keras.layers.Flatten(data_format=data_format);
    return layer;



@accepts(dict, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_fully_connected(params, network_initializer):
    '''
    Append fc (dense) to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: fc (dense) layer
    '''
    units=params["units"];
    activation=None;
    use_bias = params["use_bias"];
    kernel_initializer=get_initializer(network_initializer);
    bias_initializer='zeros';
    kernel_regularizer=None;
    bias_regularizer=None;
    activity_regularizer=None;
    kernel_constraint=None;
    bias_constraint=None;


    layer = keras.layers.Dense(units, activation=activation, use_bias=use_bias, 
                   kernel_initializer=kernel_initializer, bias_initializer=bias_initializer, 
                   kernel_regularizer=kernel_regularizer, bias_regularizer=bias_regularizer, 
                   activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint, 
                   bias_constraint=bias_constraint)

    
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_dropout(params):
    '''
    Append dropout to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: dropout layer
    '''
    rate=params["drop_probability"];
    layer = keras.layers.Dropout(rate, noise_shape=None, seed=None);

    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_identity(params):
    '''
    Append idenity to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: idenity layer
    '''
    layer = keras.activations.linear;
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_batch_normalization(params):
    '''
    Append batchnorm to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural netwrk layer: batchnorm layer
    '''
    axis = -1;
    momentum=params["moving_average_momentum"];
    epsilon=params["epsilon"];
    center=params["use_trainable_parameters"];
    scale=params["use_trainable_parameters"];
    beta_initializer='zeros';
    gamma_initializer='ones';
    moving_mean_initializer='zeros';
    moving_variance_initializer='ones';
    beta_regularizer=None;
    gamma_regularizer=None;
    beta_constraint=None;
    gamma_constraint=None;



    layer = keras.layers.BatchNormalization(axis=axis, momentum=momentum, epsilon=epsilon, 
                                center=center, scale=scale, beta_initializer=beta_initializer, 
                                gamma_initializer=gamma_initializer, 
                                moving_mean_initializer=moving_mean_initializer, 
                                moving_variance_initializer=moving_variance_initializer, 
                                beta_regularizer=beta_regularizer, gamma_regularizer=gamma_regularizer, 
                                beta_constraint=beta_constraint, gamma_constraint=gamma_constraint);


    return layer;



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_relu(params):
    '''
    Append relu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: relu activation
    '''
    layer = keras.layers.ReLU();
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softmax(params):
    '''
    Append sigmoid to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: sigmoid activation
    '''
    axis = params["axis"];

    layer = keras.layers.Softmax(axis=axis);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_thresholded_relu(params):
    '''
    Append thresholded-relu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: thresholded-reli activation
    '''
    threshold = params["threshold"];

    layer = keras.layers.ThresholdedReLU(theta=threshold);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_elu(params):
    '''
    Append elu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: elu activation
    '''
    alpha = params["alpha"];
    
    layer = keras.layers.ELU(alpha=alpha);
    return layer



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_prelu(params):
    '''
    Append prelu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: prelu activation
    '''
    layer = keras.layers.PReLU();
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_leaky_relu(params):
    '''
    Append leaky-relu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: leaky-relu activation
    '''
    alpha = params["alpha"];
    
    layer = keras.layers.LeakyReLU(alpha=alpha);
    return layer



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_selu(params):
    '''
    Append selu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: selu activation
    '''
    layer = keras.layers.Activation('selu');
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softplus(params):
    '''
    Append softplus to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softplus activation
    '''
    layer =keras.layers.Activation('softplus');
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softsign(params):
    '''
    Append softsign to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softsign activation
    '''
    layer = keras.layers.Activation('softsign');
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_tanh(params):
    '''
    Append tanh to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: tanh activation
    '''
    layer = keras.layers.Activation('tanh');
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_sigmoid(params):
    '''
    Append sigmoid to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: sigmoid activation
    '''
    layer = keras.layers.Activation('sigmoid');
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_hard_sigmoid(params):
    '''
    Append hard-sigmoid to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: hard-sigmoid activation
    '''
    layer = keras.layers.Activation('hard_sigmoid');
    return layer





from monk.tf_keras_1.models.imports import *
from monk.system.imports import *
from monk.tf_keras_1.models.models import *
from monk.tf_keras_1.models.common import create_final_layer
from monk.tf_keras_1.models.common import get_layer_uid
from monk.tf_keras_1.models.layers import custom_model_get_layer



@accepts(dict, path=[str, bool], final=bool, resume=bool, external_path=[bool, str, list], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def load_model(system_dict, path=False, final=False, resume=False, external_path=False):
    '''
    Load model based on the system state 

    Args:
        system_dict (dict): System Dictionary
        path (str): Path to final or best model weights if Final flag is set
        final (bool): If True, Load model generated from latest epoch training
        resume (bool): If True, load model from last checkpoint to resume training
        external_path (str): Path to custom model weights

    Returns:
        network: Neural network loaded with weights.
    '''
    if(final):
        if(path):
            finetune_net = keras.models.load_model(path + "final.h5");
        else:
            finetune_net = keras.models.load_model(system_dict["model_dir_relative"] + "final.h5");
    if(resume):
        finetune_net = keras.models.load_model(system_dict["model_dir_relative"] + "resume_state.h5");
 
    if(external_path):
        finetune_net = keras.models.load_model(external_path);

    return finetune_net;






@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def setup_model(system_dict):
    '''
    Setup model based on the system state and parameters

    Args:
        system_dict (dict): System Dictionary

    Returns:
        dict: Updated system dictionary
    '''
    if(system_dict["model"]["type"] == "pretrained"):
        model_name = system_dict["model"]["params"]["model_name"];
        use_pretrained = system_dict["model"]["params"]["use_pretrained"];
        freeze_base_network = system_dict["model"]["params"]["freeze_base_network"];
        custom_network = system_dict["model"]["custom_network"];
        final_layer = system_dict["model"]["final_layer"];
        num_classes = system_dict["dataset"]["params"]["num_classes"];
        input_size = system_dict["dataset"]["params"]["input_size"];

        finetune_net, model_name = get_base_model(model_name, use_pretrained, num_classes, freeze_base_network, input_size);


        if(len(custom_network)):
            if(final_layer):
                finetune_net = create_final_layer(finetune_net, custom_network, num_classes);
            else:
                msg = "Final layer not assigned";
                raise ConstraintError(msg);
        else:
            x = finetune_net.output
            x = krl.GlobalAveragePooling2D()(x)
            x = krl.Dense(512)(x);
            x = krl.ReLU()(x)
            x = krl.Dropout(0.5)(x);
            if(system_dict["dataset"]["label_type"] == "single" or system_dict["dataset"]["label_type"] == False):
                x = krl.Dense(num_classes)(x);
                preds = krl.Softmax()(x);
            else:
                x = krl.Dense(num_classes)(x);
                preds = krl.Activation('sigmoid')(x);
            finetune_net = keras.models.Model(inputs=finetune_net.input, outputs=preds);

        system_dict["local"]["model"] = finetune_net;

        return system_dict;


    else:
        net = create_network(system_dict["custom_model"]["network_stack"], 
            system_dict["dataset"]["params"]["data_shape"], 
            system_dict["custom_model"]["network_initializer"]);

        system_dict["local"]["model"] = net;

        return system_dict;





#@TraceFunction(trace_args=True, trace_rv=False)
def create_block(network_stack, count, G, sequential_first, position, current_width, out, initializer):
    '''
    Recursively create sub-blocks when designing custom networks

    Args:
        network_stack (list): List of lists containing information on layers for the sub-branch in the network
        count (dict): A dictionary mapping to a count of every type of layer in the network
        G (directed graph): NetworkX object
        sequential_first (str): NAme of the current input layer
        position (int): Vertical position on the directed graph
        current_width (int): Horizontal position on the directed graph
        out (placeholder): Keras placeholder temporary input for this sub-network
        initializer (keras iniatialzer): Random wieght initializer for layers in the sub-network
    
    Returns:
        neural network: The required sub-branch
        directed graph: Updated directed graph
        str: Name of the outermost layer in the sub-network
        int: Vertical position of the outer most layer in the sub-network 
        int: Horizontal position of the outer most layer in the sub-network 
    '''
    position += 1;
    max_width = current_width;
    for i in range(len(network_stack)): 
        if(type(network_stack[i]) == list):
            is_block = True;

            if(type(network_stack[i][-1]) != list):
                if(network_stack[i][-1]["name"] in ["add", "concatenate"]):
                    is_block=False;

            if(is_block):
                G, count, sequential_second, position, _, out, initializer =  create_block(network_stack[i], count,
                                                                                        G, sequential_first, position, 0,
                                                                                        out, initializer)
                #print(sequential_second, current_in_shape)
                sequential_first = sequential_second
            else:
                branch_end_points = [];
                branch_max_length = 0;
                branches = [];
                branch_net = [];
                out_base = out;

                #if(max_width < len(network_stack[i])-2):
                #    max_width = len(network_stack[i])-2;
                max_width = current_width;
                width = current_width;
                for j in range(len(network_stack[i])-1):
                    small_net = [];
                    branch_net.append(out_base)
                    branch_first = sequential_first
                    branch_position = position
                    column = max((j+1)*2+current_width, width);
                    max_width = column;
                    for k in range(len(network_stack[i][j])):
                        if type(network_stack[i][j][k]) == list:
                            is_block2 = True;

                            if(type(network_stack[i][j][k][-1]) != list):
                                if(network_stack[i][j][k][-1]["name"] in ["add", "concatenate"]):
                                    is_block2=False;

                            if(is_block2):
                                G, count, branch_second, branch_position, width, branch_net[j], initializer = create_block(network_stack[i][j][k], 
                                                                                                             count,
                                                                                                             G, 
                                                                                                             branch_first, 
                                                                                                             branch_position, 
                                                                                                             column-2, #j+k+width
                                                                                                             branch_net[j], 
                                                                                                             initializer)
                            else:
                                G, count, branch_second, branch_position, width, branch_net[j], initializer = create_block([network_stack[i][j][k]], 
                                                                                                             count,
                                                                                                             G, 
                                                                                                             branch_first, 
                                                                                                             branch_position, 
                                                                                                             column-2,
                                                                                                             branch_net[j],
                                                                                                             initializer)
                            #print(branch_second, current_in_shape_base)
                            branch_first = branch_second
                            small_net.append(branch_net[j]);
                        else:
                            branch_second, count = get_layer_uid(network_stack[i][j][k], count);
                            branch_net[j] = custom_model_get_layer(network_stack[i][j][k], initializer)(branch_net[j]);
                            small_net.append(custom_model_get_layer(network_stack[i][j][k], initializer));
                            G.add_node(branch_second, pos=(column, branch_position));
                            branch_position += 1;
                            G.add_edge(branch_first, branch_second);
                            branch_first = branch_second;

                        branch_max_length = max(branch_position, branch_max_length)
                        if(k == len(network_stack[i][j])-1):
                            branch_end_points.append(branch_second);
                    branches.append(small_net);

                position = branch_max_length;
                position += 1;
                max_width += 2;

                sequential_second, count = get_layer_uid(network_stack[i][-1], count);
                if(network_stack[i][-1]["name"] == "concatenate"):
                    out = keras.layers.concatenate(branch_net)
                else:
                    out = keras.layers.add(branch_net)

                G.add_node(sequential_second, pos=(2 + current_width, position));
                position += 1;
                for i in range(len(branch_end_points)):
                    G.add_edge(branch_end_points[i], sequential_second);
                sequential_first = sequential_second;
        else:
            sequential_second, count = get_layer_uid(network_stack[i], count);
            out = custom_model_get_layer(network_stack[i], initializer)(out);
            G.add_node(sequential_second, pos=(2 + current_width, position))
            position += 1;
            G.add_edge(sequential_first, sequential_second);
            sequential_first = sequential_second;
            
    return G, count, sequential_second, position, max_width, out, initializer
    


#@TraceFunction(trace_args=True, trace_rv=False)
def create_network(network_stack, current_in_shape, initializer):
    '''
    Main function to create network when designing custom networks

    Args:
        network_stack (list): List of lists containing information on layers in the network
        initializer (keras iniatialzer): Random wieght initializer for layers in the sub-network
    
    Returns:
        neural network: The required complete network
    '''
    count = [];
    for i in range(len(names)):
        count.append(1);

    G=nx.DiGraph()
    G.add_node("Net", pos=(1,1))
    sequential_first = "data";
    

    if(len(current_in_shape) == 2):
        c, w = current_in_shape;
        input1 = krl.Input(shape=(w, c));
    elif(len(current_in_shape) == 3):
        c, h, w = current_in_shape;
        input1 = krl.Input(shape=(h, w, c));
    elif(len(current_in_shape) == 4):
        c, d, h, w = current_in_shape;
        input1 = krl.Input(shape=(d, h, w, c));

    out = input1;

    count = [];
    for i in range(len(names)):
        count.append(1);

    position = 1;
    G.add_node(sequential_first, pos=(2,1))
    position += 1;


    max_width = 1;
    width = 0;
    for i in range(len(network_stack)):
        if(type(network_stack[i]) == list):
            is_block = True;

            if(type(network_stack[i][-1]) != list):
                if(network_stack[i][-1]["name"] in ["add", "concatenate"]):
                    is_block=False;


            if(is_block):
                G, count, sequential_second, position, _, out, initializer =  create_block(network_stack[i], count,
                                                                                        G, sequential_first, position, 0,
                                                                                        out, initializer)
                #print(sequential_second, current_in_shape)
                sequential_first = sequential_second
            else:
                branch_end_points = [];
                branch_max_length = 0;
                branches = [];
                branch_net = [];
                out_base = out;


                if(max_width < len(network_stack[i])-2):
                    max_width = len(network_stack[i])-2;
                width = 0;
                for j in range(len(network_stack[i])-1):
                    small_net = [];
                    branch_first = sequential_first
                    branch_net.append(out_base)
                    branch_position = position
                    if(width > 0):
                        if(column == width):
                            column += 2;
                        else:
                            column = width;
                    else:
                        column = (j+1)*2;
                    for k in range(len(network_stack[i][j])):
                        if type(network_stack[i][j][k]) == list:
                            is_block2 = True;

                            if(type(network_stack[i][j][k][-1]) != list):
                                if(network_stack[i][j][k][-1]["name"] in ["add", "concatenate"]):
                                    is_block2=False;
                        
                            if(is_block2):
                                G, count, branch_second, branch_position, width, branch_net[j], initializer = create_block(network_stack[i][j][k], 
                                                                                                             count,
                                                                                                             G, 
                                                                                                             branch_first, 
                                                                                                             branch_position, 
                                                                                                             column-2, #j+k+width
                                                                                                             branch_net[j],
                                                                                                             initializer)
                            else:
                                G, count, branch_second, branch_position, width, branch_net[j], initializer = create_block([network_stack[i][j][k]], 
                                                                                                             count,
                                                                                                             G, 
                                                                                                             branch_first, 
                                                                                                             branch_position, 
                                                                                                             column-2,
                                                                                                             branch_net[j],
                                                                                                             initializer)
                            #print(branch_second, current_in_shape_base)
                            branch_first = branch_second
                            small_net.append(branch_net[j]);
                        else:
                            branch_second, count = get_layer_uid(network_stack[i][j][k], count);
                            branch_net[j] = custom_model_get_layer(network_stack[i][j][k], initializer)(branch_net[j]);
                            small_net.append(custom_model_get_layer(network_stack[i][j][k], initializer));
                            G.add_node(branch_second, pos=(column, branch_position));
                            branch_position += 1;
                            G.add_edge(branch_first, branch_second);
                            branch_first = branch_second;                        

                        branch_max_length = max(branch_position, branch_max_length)
                        if(k == len(network_stack[i][j])-1):
                            branch_end_points.append(branch_second);
                    branches.append(small_net);
                
                position = branch_max_length;
                position += 1;
                max_width += width

                sequential_second, count = get_layer_uid(network_stack[i][-1], count)
                if(network_stack[i][-1]["name"] == "concatenate"):
                    out = keras.layers.concatenate(branch_net)
                else:
                    out = keras.layers.add(branch_net)

                sequential_second, count = get_layer_uid(network_stack[i][-1], count);

                G.add_node(sequential_second, pos=(2, position));
                position += 1;
                for i in range(len(branch_end_points)):
                    G.add_edge(branch_end_points[i], sequential_second);
                sequential_first = sequential_second;


        else:
            sequential_second, count = get_layer_uid(network_stack[i], count);
            G.add_node(sequential_second, pos=(2, position))
            out = custom_model_get_layer(network_stack[i], initializer)(out);
            position += 1;
            G.add_edge(sequential_first, sequential_second);
            sequential_first = sequential_second;


    net = keras.models.Model(inputs=input1, outputs=out)

    max_width = max(max_width, width);
    if(max_width == 1):
        G.add_node("monk", pos=(3, position));
    else:
        G.add_node("monk", pos=(max_width + 3, position))
    pos = nx.get_node_attributes(G, 'pos')


    plt.figure(3, figsize=(12, 12 + position//6)) 
    nx.draw_networkx(G, pos, with_label=True, font_size=16, node_color="yellow", node_size=100)
    plt.savefig("graph.png");

    return net;











#@TraceFunction(trace_args=True, trace_rv=False)
def debug_create_block(network_stack, count, G, sequential_first, position, current_width):
    '''
    Recursively visualize sub-blocks when designing custom networks

    Args:
        network_stack (list): List of lists containing information on layers for the sub-branch in the network
        count (dict): A dictionary mapping to a count of every type of layer in the network
        G (directed graph): NetworkX object
        sequential_first (str): NAme of the current input layer
        position (int): Vertical position on the directed graph
        current_width (int): Horizontal position on the directed graph
        out (placeholder): Keras placeholder temporary input for this sub-network
        initializer (keras iniatialzer): Random wieght initializer for layers in the sub-network
    
    Returns:
        neural network: The required sub-branch
        directed graph: Updated directed graph
        str: Name of the outermost layer in the sub-network
        int: Vertical position of the outer most layer in the sub-network 
        int: Horizontal position of the outer most layer in the sub-network 
    '''
    position += 1;
    max_width = current_width
    for i in range(len(network_stack)): 
        if(type(network_stack[i]) == list):
            is_block = True;

            if(type(network_stack[i][-1]) != list):
                if(network_stack[i][-1]["name"] in ["add", "concatenate"]):
                    is_block=False;

            if(is_block):
                G, count, sequential_second, position, _ =  debug_create_block(network_stack[i], count,
                                                                                G, sequential_first, position, current_width) #0
                sequential_first = sequential_second
            else:
                branch_end_points = [];
                branch_max_length = 0;
                branches = [];
                branch_net = [];

                #if(max_width < len(network_stack[i])-2):
                #    max_width = len(network_stack[i])-2;
                max_width = current_width;
                width = current_width;
                for j in range(len(network_stack[i])-1):
                    branch_first = sequential_first
                    branch_position = position
                    column = max((j+1)*2+current_width, width);
                    max_width = column
                    for k in range(len(network_stack[i][j])):
                        if type(network_stack[i][j][k]) == list:
                            is_block2 = True;

                            if(type(network_stack[i][j][k][-1]) != list):
                                if(network_stack[i][j][k][-1]["name"] in ["add", "concatenate"]):
                                    is_block2=False;


                            if(is_block2):
                                G, count, branch_second, branch_position, width = debug_create_block(network_stack[i][j][k], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width, j*2+current_width
                            else:
                                G, count, branch_second, branch_position, width = debug_create_block([network_stack[i][j][k]], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width, j+k+current_width
                            branch_first = branch_second
                        else:
                            branch_second, count = get_layer_uid(network_stack[i][j][k], count);
                            G.add_node(branch_second, pos=(column, branch_position));
                            branch_position += 1;
                            G.add_edge(branch_first, branch_second);
                            branch_first = branch_second;

                        branch_max_length = max(branch_position, branch_max_length)
                        if(k == len(network_stack[i][j])-1):
                            branch_end_points.append(branch_second);
                
                position = branch_max_length;
                position += 1;
                max_width += 2;

                sequential_second, count = get_layer_uid(network_stack[i][-1], count);

                G.add_node(sequential_second, pos=(2 + current_width, position));
                position += 1;
                for i in range(len(branch_end_points)):
                    G.add_edge(branch_end_points[i], sequential_second);
                sequential_first = sequential_second;
        else:
            sequential_second, count = get_layer_uid(network_stack[i], count);
            G.add_node(sequential_second, pos=(2+current_width, position))
            position += 1;
            G.add_edge(sequential_first, sequential_second);
            sequential_first = sequential_second;
    
    return G, count, sequential_second, position, max_width
    

#@TraceFunction(trace_args=True, trace_rv=False)
def debug_create_network(network_stack):
    '''
    Main function to visualize network when designing custom networks

    Args:
        network_stack (list): List of lists containing information on layers in the network
        initializer (keras iniatialzer): Random wieght initializer for layers in the sub-network
    
    Returns:
        neural network: The required complete network
    '''
    count = [];
    for i in range(len(names)):
        count.append(1);

    G=nx.DiGraph()
    G.add_node("Net", pos=(1,1))
    sequential_first = "data";
    #sequential_second, count = get_layer_uid(network_stack[0], count);

    count = [];
    for i in range(len(names)):
        count.append(1);

    position = 1;
    G.add_node(sequential_first, pos=(2,1))
    position += 1;


    max_width = 1;
    width = 0;
    for i in range(len(network_stack)):
        if(type(network_stack[i]) == list):
            is_block = True;

            if(type(network_stack[i][-1]) != list):
                if(network_stack[i][-1]["name"] in ["add", "concatenate"]):
                    is_block=False;

            if(is_block):
                G, count, sequential_second, position, _ =  debug_create_block(network_stack[i], count,
                                                                                G, sequential_first, position, 0)
                sequential_first = sequential_second
            else:
                branch_end_points = [];
                branch_max_length = 0;
                branches = [];
                branch_net = [];


                if(max_width < len(network_stack[i])-2):
                    max_width = len(network_stack[i])-2;
                width = 0;
                for j in range(len(network_stack[i])-1):
                    branch_first = sequential_first
                    branch_position = position
                    if(width > 0):
                        if(column == width):
                            column += 2;
                        else:
                            column = width;
                    else:
                        column = (j+1)*2;
                    for k in range(len(network_stack[i][j])):
                        if type(network_stack[i][j][k]) == list:
                            is_block2 = True;

                            if(type(network_stack[i][j][k][-1]) != list):
                                if(network_stack[i][j][k][-1]["name"] in ["add", "concatenate"]):
                                    is_block2=False;

                           

                            if(is_block2):
                                G, count, branch_second, branch_position, width = debug_create_block(network_stack[i][j][k], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j*2+width
                            else:
                                G, count, branch_second, branch_position, width = debug_create_block([network_stack[i][j][k]], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width
                            branch_first = branch_second
                        else:
                            branch_second, count = get_layer_uid(network_stack[i][j][k], count);
                            G.add_node(branch_second, pos=(column, branch_position));
                            branch_position += 1;
                            G.add_edge(branch_first, branch_second);
                            branch_first = branch_second;

                        branch_max_length = max(branch_position, branch_max_length)
                        if(k == len(network_stack[i][j])-1):
                            branch_end_points.append(branch_second);
                    
                position = branch_max_length;
                position += 1;
                max_width += width

                sequential_second, count = get_layer_uid(network_stack[i][-1], count);

                G.add_node(sequential_second, pos=(2, position));
                position += 1;
                for i in range(len(branch_end_points)):
                    G.add_edge(branch_end_points[i], sequential_second);
                sequential_first = sequential_second;


        else:
            sequential_second, count = get_layer_uid(network_stack[i], count);
            G.add_node(sequential_second, pos=(2, position))
            position += 1;
            G.add_edge(sequential_first, sequential_second);
            sequential_first = sequential_second;

    max_width = max(max_width, width);
    if(max_width == 1):
        G.add_node("monk", pos=(3, position));
    else:
        G.add_node("monk", pos=(max_width + 3, position))
    pos = nx.get_node_attributes(G, 'pos')


    plt.figure(3, figsize=(16, 20 + position//6)) 
    nx.draw_networkx(G, pos, with_label=True, font_size=16, node_color="yellow", node_size=100)
    plt.savefig("graph.png");


from monk.gluon.testing.imports import *
from monk.system.imports import *

@accepts(str, bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def process_single(img_name, return_raw, system_dict):
    '''
    Run inference on a single image

    Args:
        img_name (str): path to image
        return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                            Else, only the most probable class score is returned back.
                            

    Returns:
        str: predicted class
        float: prediction score
    '''
    img = image.imread(img_name)
    img = system_dict["local"]["data_transforms"]["test"](img).expand_dims(axis=0);
    img = img.copyto(system_dict["local"]["ctx"][0]);
    outputs = system_dict["local"]["model"](img);
    if(system_dict["dataset"]["params"]["classes"]):
        prediction = system_dict["dataset"]["params"]["classes"][np.argmax(outputs[0].asnumpy())];
    else:
        prediction = np.argmax(outputs[0].asnumpy());

    normalized = softmax(outputs[0].asnumpy())
    score = normalized[np.argmax(normalized)];
    if(return_raw):
        return prediction, score, normalized
    else:
        return prediction, score, "";
    



@accepts(str, bool, float, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def process_multi(img_name, return_raw, img_thresh, system_dict):
    '''
    Run inference on a single image when label type is multi-label

    Args:
        img_name (str): path to image
        return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                            Else, only the most probable class score is returned back.
        img_thresh (float): Thresholding for multi label image classification.
                            

    Returns:
        list: list of predicted classes
        list: list of predicted scores
    '''

    img = image.imread(img_name)
    img = system_dict["local"]["data_transforms"]["test"](img).expand_dims(axis=0);
    img = img.copyto(system_dict["local"]["ctx"][0]);
    outputs = system_dict["local"]["model"](img);

    list_classes = [];
    raw_scores = outputs[0].asnumpy();
    list_scores = [];


    for i in range(len(raw_scores)):
        prob = logistic.cdf(raw_scores[i])
        if(prob > img_thresh):
            list_classes.append(system_dict["dataset"]["params"]["classes"][i])
            list_scores.append(prob)


    if(return_raw):
        return list_classes, list_scores, raw_scores;
    else:
        return list_classes, list_scores, "";



from monk.gluon.finetune.imports import *
from monk.system.imports import *


from monk.gluon.finetune.level_6_params_main import prototype_params



class prototype_aux(prototype_params):
    '''
    Main class for all auxiliary functions - EDA, Estimate Training Time, Resetting params, switching modes, & debugging 

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);

    ###############################################################################################################################################
    def EDA(self, show_img=False, save_img=False, check_missing=False, check_corrupt=False):
        '''
        Experimental Data Analysis
            - Finding number of images in each class
            - Check missing images in case of csv type dataset
            - Find all corrupt images
        Args:
            show_img (bool): If True, displays bar graph for images per class 
            save_img (bool): If True, saves bar graph for images per class
            check_missing (bool): If True, checks for missing images in csv type dataset
            check_corrupt (bool): If True, checks for corrupted images in foldered and csv dataset

        Returns:
            None
        '''
        if(not self.system_dict["dataset"]["train_path"]):
            msg = "Dataset train path not set. Cannot run EDA";
            raise ConstraintError(msg);


        classes_folder, classes_folder_strength = class_imbalance(self.system_dict, show_img, save_img);
        missing_images_train, missing_images_val, corrupt_images_train, corrupt_images_val = corrupted_missing_images(self.system_dict, check_missing, check_corrupt);

        self.custom_print("EDA: Class imbalance")
        for i in range(len(classes_folder)):
            self.custom_print("    {}. Class: {}, Number: {}".format(i+1, classes_folder[i], classes_folder_strength[i]));
        self.custom_print("");

        if(check_missing):
            self.custom_print("EDA: Check Missing");
            if("csv" in self.system_dict["dataset"]["dataset_type"]):
                if(missing_images_train):
                    self.custom_print("    Missing Images in folder {}".format(self.system_dict["dataset"]["train_path"]));
                    for i in range(len(missing_images_train)):
                        self.custom_print("        {}. {}".format(i+1, missing_images_train[i]));
                    self.custom_print("");
                else:
                    self.custom_print("    All images present in train dir.");
                    self.custom_print("");

                if(missing_images_val):
                    self.custom_print("    Missing Images in folder {}".format(self.system_dict["dataset"]["val_path"]));
                    for i in range(len(missing_images_val)):
                        self.custom_print("        {}. {}".format(i+1, missing_images_val[i]));
                    self.custom_print("");
                else:
                    self.custom_print("    All images present in val dir.");
                    self.custom_print("");
            else:
                self.custom_print("    Missing check not required for foldered dataset");
                self.custom_print("");
            

        if(check_corrupt):
            self.custom_print("EDA: Check Corrupt");
            if(corrupt_images_train):
                self.custom_print("    Corrupt Images in folder {}".format(self.system_dict["dataset"]["train_path"]));
                for i in range(len(corrupt_images_train)):
                    self.custom_print("        {}. {}".format(i+1, corrupt_images_train[i]));
                self.custom_print("");
            else:
                self.custom_print("    No corrupt image found in train dir.");
                self.custom_print("");

            if(corrupt_images_val):
                self.custom_print("    Corrupt Images in folder {}".format(self.system_dict["dataset"]["val_path"]));
                for i in range(len(corrupt_images_val)):
                    self.custom_print("        {}. {}".format(i+1, corrupt_images_val[i]));
                self.custom_print("");
            else:
                self.custom_print("    No corrupt image found in val dir.");
                self.custom_print("");
    ###############################################################################################################################################



    ###############################################################################################################################################
    def Freeze_Layers(self, num=10):
        '''
        Freeze first "n" trainable layers in the network
        Args:
            num (int): Number of layers to freeze

        Returns:
            None
        '''
        self.num_freeze = num;
        self.system_dict = freeze_layers(num, self.system_dict);
        self.custom_print("Model params post freezing");
        self.custom_print("    Num trainable layers: {}".format(self.system_dict["model"]["params"]["num_params_to_update"]));
        self.custom_print("");
        save(self.system_dict);
    ###############################################################################################################################################



    ###############################################################################################################################################
    def Estimate_Train_Time(self, num_epochs=False):
        '''
        Estimate training time before running training


        Args:
            num_epochs (int): Number of epochs to be trained and get eestimation for it.

        Returns:
            None
        '''
        total_time_per_epoch = self.get_training_estimate();
        self.custom_print("Training time estimate");
        if(not num_epochs):
            total_time = total_time_per_epoch*self.system_dict["hyper-parameters"]["num_epochs"];
            self.custom_print("    {} Epochs: Approx. {} Min".format(self.system_dict["hyper-parameters"]["num_epochs"], int(total_time//60)+1));
            self.custom_print("");
        else:
            total_time = total_time_per_epoch*num_epochs;
            self.custom_print("    {} Epochs: Approx. {} Min".format(num_epochs, int(total_time//60)+1));
            self.custom_print("");
    ###############################################################################################################################################


    ##########################################################################################################################################################
    def Reload(self):
        '''
        Function to actuate all the updates in the update and expert modes


        Args:
            None

        Returns:
            None
        '''
        if(self.system_dict["states"]["eval_infer"]):
            del self.system_dict["local"]["data_loaders"];
            self.system_dict["local"]["data_loaders"] = {};
            self.Dataset();
            del self.system_dict["local"]["model"];
            self.system_dict["local"]["model"] = False;
            self.Model();
        else:
            if(not self.system_dict["states"]["copy_from"]):
                self.system_dict["local"]["model"].collect_params().reset_ctx([mx.cpu()]);
                del self.system_dict["local"]["model"];
                self.system_dict["local"]["model"] = False;
            del self.system_dict["local"]["data_loaders"];
            self.system_dict["local"]["data_loaders"] = {};
            self.Dataset();
            if(not self.system_dict["states"]["copy_from"]):
                self.Model();
                self.system_dict = load_scheduler(self.system_dict);
                self.system_dict = load_optimizer(self.system_dict);
                self.system_dict = load_loss(self.system_dict);
            if(self.system_dict["model"]["params"]["num_freeze"]):
                self.system_dict = freeze_layers(self.system_dict["model"]["params"]["num_freeze"], self.system_dict);
                self.custom_print("Model params post freezing");
                self.custom_print("    Num trainable layers: {}".format(self.system_dict["model"]["params"]["num_params_to_update"]));
                self.custom_print("");
                save(self.system_dict);
    ##########################################################################################################################################################

    
    
    
    


    ##########################################################################################################################################################
    def reset_transforms(self, test=False):
        '''
        Reset transforms to change them.


        Args:
            test (bool): If True, test transforms are reset,
                          Else, train and validation transforms are reset.

        Returns:
            None
        '''
        if(self.system_dict["states"]["eval_infer"] or test):
            self.system_dict["local"]["transforms_test"] = [];
            self.system_dict["local"]["normalize"] = False;
            self.system_dict["dataset"]["transforms"]["test"] = [];
        else:
            self.system_dict["local"]["transforms_train"] = [];
            self.system_dict["local"]["transforms_val"] = [];
            self.system_dict["local"]["normalize"] = False;
            self.system_dict["dataset"]["transforms"]["train"] = [];
            self.system_dict["dataset"]["transforms"]["val"] = [];
        save(self.system_dict);
    ##########################################################################################################################################################

    


    ##########################################################################################################################################################
    def reset_model(self):
        '''
        Reset model to update and reload it with custom weights.


        Args:
            None

        Returns:
            None
        '''
        if(self.system_dict["states"]["copy_from"]):
            msg = "Cannot reset model in Copy-From mode.\n";
            raise ConstraintError(msg)
        self.system_dict["model"]["custom_network"] = [];
        self.system_dict["model"]["final_layer"] = None;
    ##########################################################################################################################################################



    ##########################################################################################################################################################
    def Switch_Mode(self, train=False, eval_infer=False):
        '''
        Switch modes between training an inference without reloading the experiment


        Args:
            train (bool): If True, switches to training mode
            eval_infer (bool): If True, switches to validation and inferencing mode

        Returns:
            None
        '''
        if(eval_infer):
            self.system_dict["states"]["eval_infer"] = True;
        elif(train):
            self.system_dict["states"]["eval_infer"] = False;
    ##########################################################################################################################################################




    ##########################################################################################################################################################
    def debug_custom_model_design(self, network_list):
        '''
        Debug model while creating it. 
        Saves image as graph.png which is displayed 


        Args:
            network_list (list): List containing network design

        Returns:
            None
        '''
        debug_create_network(network_list);
        if(not isnotebook()):
            self.custom_print("If not using notebooks check file generated graph.png");


    ##########################################################################################################################################################


    ##########################################################################################################################################################
    def Visualize_Kernels(self, store_images_if_notebook=False):
        '''
        Visualize kernel weights of model

        Args:
            store_images_if_notebook (bool): If the images need to be stored instead of IPython widget 
                                             while using notebook. Not applicable for other environments.
        Returns:
            IPython widget displaying kernel weights if used inside a notebook.
            Else stores the maps in the visualization directory. 
        '''
        is_notebook = isnotebook()

        visualizer = CNNVisualizer(self.system_dict["local"]["model"], is_notebook)
        
        if(not is_notebook) or (store_images_if_notebook):
            self.custom_print("The images will be stored in the visualization directory of the experiment");
            
            from system.common import create_dir
            create_dir(self.system_dict["visualization"]["base"])
            create_dir(self.system_dict["visualization"]["kernels_dir"])

            visualizer.visualize_kernels(self.system_dict["visualization"]["kernels_dir"])

        else:
            visualizer.visualize_kernels()
    ##########################################################################################################################################################


    ##########################################################################################################################################################
    def Visualize_Feature_Maps(self, image_path, store_images_if_notebook=False):
        '''
        Visualize feature maps generated by model on an image

        Args:
            image_path (str): Path to the image
            store_images_if_notebook (bool): If the images need to be stored instead of IPython widget 
                                             while using notebook. Not applicable for other environments.
        
        Returns:
            IPython widget displaying feature maps if used inside a notebook.
            Else stores the maps in the visualization directory.  
        '''
        is_notebook = isnotebook()

        visualizer = CNNVisualizer(self.system_dict["local"]["model"], is_notebook)

        if(self.system_dict["model"]["params"]["use_gpu"]):
            ctx = mx.gpu() if mx.context.num_gpus() else mx.cpu()
        else:
            ctx = mx.cpu()

        if(not is_notebook) or (store_images_if_notebook):
            self.custom_print("The images will be stored in the visualization directory of the experiment");

            from system.common import create_dir
            create_dir(self.system_dict["visualization"]["base"])
            create_dir(self.system_dict["visualization"]["feature_maps_dir"])
            img_name = "".join(image_path.split("/")[-1].split(".")[0:-1])
            img_dir = self.system_dict["visualization"]["feature_maps_dir"] + img_name + '/'
            create_dir(img_dir)
            
            visualizer.visualize_feature_maps(image_path, ctx=ctx, store_path=img_dir)

        else:
            visualizer.visualize_feature_maps(image_path, ctx=ctx)


from monk.tf_keras_1.losses.imports import *
from monk.system.imports import *


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def retrieve_loss(system_dict):
    '''
    Retrieve loss post state changes

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = system_dict["hyper-parameters"]["loss"]["name"];
    name = system_dict["local"]["criterion"];

    if(name == "l1"):
        system_dict["local"]["criterion"] = krlo.mean_absolute_error

    elif(name == "l2"):
        system_dict["local"]["criterion"] = krlo.mean_squared_error

    elif(name == "crossentropy"):
        system_dict["local"]["criterion"] = krlo.categorical_crossentropy;

    elif(name == "binarycrossentropy"):
        system_dict["local"]["criterion"] = krlo.binary_crossentropy;

    elif(name == "kldiv"):
        system_dict["local"]["criterion"] = krlo.kullback_leibler_divergence;

    elif(name == "hinge"):
        system_dict["local"]["criterion"] = krlo.hinge;

    elif(name == "squaredhinge"):
        system_dict["local"]["criterion"] = krlo.squared_hinge;

    return system_dict;


from monk.gluon.models.imports import *
from monk.system.imports import *
from monk.gluon.models.models import *
from monk.gluon.models.common import create_final_layer
from monk.gluon.models.common import get_layer_uid
from monk.gluon.models.layers import custom_model_get_layer
from monk.gluon.models.layers import addBlock
from monk.gluon.models.initializers import initialize_network



@accepts(dict, path=[str, bool], final=bool, resume=bool, external_path=[bool, str, list], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def load_model(system_dict, path=False, final=False, resume=False, external_path=False):
    '''
    Load model based on the system state 

    Args:
        system_dict (dict): System Dictionary
        path (str): Path to final or best model weights if Final flag is set
        final (bool): If True, Load model generated from latest epoch training
        resume (bool): If True, load model from last checkpoint to resume training
        external_path (str): Path to custom model weights

    Returns:
        network: Neural network loaded with weights.
    '''
    if(final):
        if(path):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore");
                finetune_net = mx.gluon.SymbolBlock.imports(path + 'final-symbol.json', ['data'], path + 'final-0000.params');
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                finetune_net = mx.gluon.SymbolBlock.imports(system_dict["model_dir_relative"] + 'final-symbol.json', ['data'], system_dict["model_dir_relative"] + 'final-0000.params');
    if(resume):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore");
            finetune_net = mx.gluon.SymbolBlock.imports(system_dict["model_dir_relative"] + 'resume_state-symbol.json', ['data'], system_dict["model_dir_relative"] + 'resume_state-0000.params');
 
    if(external_path):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore");
            finetune_net = mx.gluon.SymbolBlock.imports(external_path[0], ['data'], external_path[1]);

    return finetune_net;



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def setup_model(system_dict):
    '''
    Setup model based on the system state and parameters

    Args:
        system_dict (dict): System Dictionary

    Returns:
        dict: Updated system dictionary
    '''
    if(system_dict["model"]["type"] == "pretrained"):
        model_name = system_dict["model"]["params"]["model_name"];
        use_pretrained = system_dict["model"]["params"]["use_pretrained"];
        freeze_base_network = system_dict["model"]["params"]["freeze_base_network"];
        custom_network = system_dict["model"]["custom_network"];
        final_layer = system_dict["model"]["final_layer"];
        num_classes = system_dict["dataset"]["params"]["num_classes"];

        finetune_net, model_name = get_base_model(model_name, use_pretrained, num_classes, freeze_base_network);

        if(len(custom_network)):
            if(final_layer):
                if(model_name in set1):
                    finetune_net = create_final_layer(finetune_net, custom_network, num_classes, set=1);
                elif(model_name in set2):
                    finetune_net = create_final_layer(finetune_net, custom_network, num_classes, set=2);
                elif(model_name in set3):
                    finetune_net = create_final_layer(finetune_net, custom_network, num_classes, set=3);
            else:
                print("Final layer not assigned");
                return 0;
        else:
            if(model_name in set1):
                with finetune_net.name_scope():
                    finetune_net.output = nn.Dense(num_classes, weight_initializer=init.Xavier());
                    finetune_net.output.initialize(init.Xavier(), ctx = ctx);
            elif(model_name in set2):
                net = nn.HybridSequential();
                with net.name_scope():
                    net.add(nn.Conv2D(num_classes, kernel_size=(1, 1), strides=(1, 1), weight_initializer=init.Xavier()));
                    net.add(nn.Flatten());
                with finetune_net.name_scope():
                    finetune_net.output = net;
                    finetune_net.output.initialize(init.Xavier(), ctx = ctx)
            elif(model_name in set3):
                with finetune_net.name_scope():
                    finetune_net.fc = nn.Dense(num_classes, weight_initializer=init.Xavier());
                    finetune_net.fc.initialize(init.Xavier(), ctx = ctx)


        if(not use_pretrained):
            finetune_net.initialize(init.Xavier(), ctx = ctx)


        system_dict["local"]["model"] = finetune_net;

        return system_dict;

    else:
        net = create_network(system_dict["custom_model"]["network_stack"], debug=system_dict["custom_model"]["debug"]);
        net = initialize_network(net, system_dict["custom_model"]["network_initializer"]);
        system_dict["local"]["model"] = net;

        return system_dict;


#@TraceFunction(trace_args=True, trace_rv=False)
def create_block(network_stack, count, G, sequential_first, position, current_width):
    '''
    Recursively create sub-blocks when designing custom networks

    Args:
        network_stack (list): List of lists containing information on layers for the sub-branch in the network
        count (dict): A dictionary mapping to a count of every type of layer in the network
        G (directed graph): NetworkX object
        sequential_first (str): NAme of the current input layer
        position (int): Vertical position on the directed graph
        current_width (int): Horizontal position on the directed graph
    
    Returns:
        neural network: The required sub-branch
        directed graph: Updated directed graph
        str: Name of the outermost layer in the sub-network
        int: Vertical position of the outer most layer in the sub-network 
        int: Horizontal position of the outer most layer in the sub-network 
    '''
    position += 1;
    max_width = current_width
    net = nn.HybridSequential();
    for i in range(len(network_stack)): 
        if(type(network_stack[i]) == list):
            is_block = True;

            if(type(network_stack[i][-1]) != list):
                if(network_stack[i][-1]["name"] in ["add", "concatenate"]):
                    is_block=False;

            if(is_block):
                block, G, count, sequential_second, position, _ =  create_block(network_stack[i], count,
                                                                                G, sequential_first, position, current_width)
                sequential_first = sequential_second
                net.add(block)
            else:
                branch_end_points = [];
                branch_max_length = 0;
                branches = [];
                branch_net = [];

                #if(max_width < len(network_stack[i])-2):
                #    max_width = len(network_stack[i])-2;
                max_width = current_width;
                width = current_width;
                for j in range(len(network_stack[i])-1):
                    small_net = [];
                    branch_net.append(nn.HybridSequential())
                    branch_first = sequential_first
                    branch_position = position
                    column = max((j+1)*2+current_width, width);
                    max_width = column
                    for k in range(len(network_stack[i][j])):
                        if type(network_stack[i][j][k]) == list:
                            is_block2 = True;

                            if(type(network_stack[i][j][k][-1]) != list):
                                if(network_stack[i][j][k][-1]["name"] in ["add", "concatenate"]):
                                    is_block2=False;

                            if(is_block2):
                                block, G, count, branch_second, branch_position, width = create_block(network_stack[i][j][k], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width
                            else:
                                block, G, count, branch_second, branch_position, width = create_block([network_stack[i][j][k]], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width
                            branch_first = branch_second
                            small_net.append(block);
                            branch_net[j].add(block);
                        else:
                            branch_second, count = get_layer_uid(network_stack[i][j][k], count);
                            small_net.append(custom_model_get_layer(network_stack[i][j][k]));
                            branch_net[j].add(custom_model_get_layer(network_stack[i][j][k]));
                            G.add_node(branch_second, pos=(column, branch_position));
                            branch_position += 1;
                            G.add_edge(branch_first, branch_second);
                            branch_first = branch_second;

                        branch_max_length = max(branch_position, branch_max_length)
                        if(k == len(network_stack[i][j])-1):
                            branch_end_points.append(branch_second);
                    branches.append(small_net);

                position = branch_max_length;
                position += 1;
                max_width += 2;

                sequential_second, count = get_layer_uid(network_stack[i][-1], count);
                if(network_stack[i][-1]["name"] == "concatenate"):
                    subnetwork = contrib_nn.HybridConcurrent(axis=1);
                    for j in range(len(network_stack[i])-1):
                        subnetwork.add(branch_net[j]);
                else:
                    subnetwork = addBlock(branches);


                G.add_node(sequential_second, pos=(2 + current_width, position));
                position += 1;
                for i in range(len(branch_end_points)):
                    G.add_edge(branch_end_points[i], sequential_second);
                sequential_first = sequential_second;
                net.add(subnetwork)
        else:
            sequential_second, count = get_layer_uid(network_stack[i], count);
            net.add(custom_model_get_layer(network_stack[i]));
            G.add_node(sequential_second, pos=(2 + current_width, position))
            position += 1;
            G.add_edge(sequential_first, sequential_second);
            sequential_first = sequential_second;
            
    return net, G, count, sequential_second, position, max_width
    


#@TraceFunction(trace_args=True, trace_rv=False)
def create_network(network_stack, debug=True):
    '''
    Main function to create network when designing custom networks

    Args:
        network_stack (list): List of lists containing information on layers in the network
    
    Returns:
        neural network: The required complete network
    '''
    count = [];
    for i in range(len(names)):
        count.append(1);

    G=nx.DiGraph()
    G.add_node("Net", pos=(1,1))
    sequential_first = "data";
    #sequential_second, count = get_layer_uid(network_stack[0], count);

    count = [];
    for i in range(len(names)):
        count.append(1);

    position = 1;
    G.add_node(sequential_first, pos=(2,1))
    position += 1;


    net = nn.HybridSequential();
    max_width = 1;
    width = 0;
    for i in range(len(network_stack)):
        if(type(network_stack[i]) == list):
            is_block = True;

            if(type(network_stack[i][-1]) != list):
                if(network_stack[i][-1]["name"] in ["add", "concatenate"]):
                    is_block=False;


            if(is_block):
                block, G, count, sequential_second, position, _ =  create_block(network_stack[i], count,
                                                                                G, sequential_first, position, 0)
                sequential_first = sequential_second
                net.add(block)
            else:
                branch_end_points = [];
                branch_max_length = 0;
                branches = [];
                branch_net = [];


                if(max_width < len(network_stack[i])-2):
                    max_width = len(network_stack[i])-2;
                width = 0;
                for j in range(len(network_stack[i])-1):
                    small_net = [];
                    branch_first = sequential_first
                    branch_net.append(nn.HybridSequential())
                    branch_position = position
                    if(width > 0):
                        if(column == width):
                            column += 2;
                        else:
                            column = width;
                    else:
                        column = (j+1)*2;
                    for k in range(len(network_stack[i][j])):
                        if type(network_stack[i][j][k]) == list:
                            is_block2 = True;

                            if(type(network_stack[i][j][k][-1]) != list):
                                if(network_stack[i][j][k][-1]["name"] in ["add", "concatenate"]):
                                    is_block2=False;
                        
                            if(is_block2):
                                block, G, count, branch_second, branch_position, width = create_block(network_stack[i][j][k], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width
                            else:
                                block, G, count, branch_second, branch_position, width = create_block([network_stack[i][j][k]], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2)
                            branch_first = branch_second
                            small_net.append(block);
                            branch_net[j].add(block);
                        else:
                            branch_second, count = get_layer_uid(network_stack[i][j][k], count);
                            small_net.append(custom_model_get_layer(network_stack[i][j][k]));
                            branch_net[j].add(custom_model_get_layer(network_stack[i][j][k]));
                            G.add_node(branch_second, pos=(column, branch_position));
                            branch_position += 1;
                            G.add_edge(branch_first, branch_second);
                            branch_first = branch_second;                        

                        branch_max_length = max(branch_position, branch_max_length)
                        if(k == len(network_stack[i][j])-1):
                            branch_end_points.append(branch_second);
                    branches.append(small_net);
                
                position = branch_max_length;
                position += 1;
                max_width += width

                sequential_second, count = get_layer_uid(network_stack[i][-1], count)
                if(network_stack[i][-1]["name"] == "concatenate"):
                    subnetwork = contrib_nn.HybridConcurrent(axis=1);
                    for j in range(len(network_stack[i])-1):
                        subnetwork.add(branch_net[j]);
                else:
                    subnetwork = addBlock(branches);

                sequential_second, count = get_layer_uid(network_stack[i][-1], count);

                G.add_node(sequential_second, pos=(2, position));
                position += 1;
                for i in range(len(branch_end_points)):
                    G.add_edge(branch_end_points[i], sequential_second);
                sequential_first = sequential_second;
                net.add(subnetwork)


        else:
            sequential_second, count = get_layer_uid(network_stack[i], count);
            G.add_node(sequential_second, pos=(2, position))
            net.add(custom_model_get_layer(network_stack[i]));
            position += 1;
            G.add_edge(sequential_first, sequential_second);
            sequential_first = sequential_second;


    max_width = max(max_width, width);
    if(max_width == 1):
        G.add_node("monk", pos=(3, position));
    else:
        G.add_node("monk", pos=(max_width + 3, position))
    pos = nx.get_node_attributes(G, 'pos')


    if(not debug):
        plt.ioff();
        plt.figure(3, figsize=(12, 12 + position//6)) 
        nx.draw_networkx(G, pos, with_label=True, font_size=16, node_color="yellow", node_size=100)
        plt.savefig("graph.png");
        plt.clf();
    else:
        plt.figure(3, figsize=(12, 12 + position//6)) 
        nx.draw_networkx(G, pos, with_label=True, font_size=16, node_color="yellow", node_size=100)
        plt.savefig("graph.png");

    return net;







#@TraceFunction(trace_args=True, trace_rv=False)
def debug_create_block(network_stack, count, G, sequential_first, position, current_width):
    '''
    Recursively visualize sub-blocks when designing custom networks 

    Args:
        network_stack (list): List of lists containing information on layers for the sub-branch in the network
        count (dict): A dictionary mapping to a count of every type of layer in the network
        G (directed graph): NetworkX object
        sequential_first (str): NAme of the current input layer
        position (int): Vertical position on the directed graph
        current_width (int): Horizontal position on the directed graph
    
    Returns:
        neural network: The required sub-branch
        directed graph: Updated directed graph
        str: Name of the outermost layer in the sub-network
        int: Vertical position of the outer most layer in the sub-network 
        int: Horizontal position of the outer most layer in the sub-network 
    '''
    position += 1;
    max_width = current_width
    for i in range(len(network_stack)): 
        if(type(network_stack[i]) == list):
            is_block = True;

            if(type(network_stack[i][-1]) != list):
                if(network_stack[i][-1]["name"] in ["add", "concatenate"]):
                    is_block=False;

            if(is_block):
                G, count, sequential_second, position, _ =  debug_create_block(network_stack[i], count,
                                                                                G, sequential_first, position, current_width) #0
                sequential_first = sequential_second
            else:
                branch_end_points = [];
                branch_max_length = 0;
                branches = [];
                branch_net = [];

                #if(max_width < len(network_stack[i])-2):
                #    max_width = len(network_stack[i])-2;
                max_width = current_width;
                width = current_width;
                for j in range(len(network_stack[i])-1):
                    branch_first = sequential_first
                    branch_position = position
                    column = max((j+1)*2+current_width, width);
                    max_width = column
                    for k in range(len(network_stack[i][j])):
                        if type(network_stack[i][j][k]) == list:
                            is_block2 = True;

                            if(type(network_stack[i][j][k][-1]) != list):
                                if(network_stack[i][j][k][-1]["name"] in ["add", "concatenate"]):
                                    is_block2=False;


                            if(is_block2):
                                G, count, branch_second, branch_position, width = debug_create_block(network_stack[i][j][k], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width, j*2+current_width
                            else:
                                G, count, branch_second, branch_position, width = debug_create_block([network_stack[i][j][k]], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width, j+k+current_width
                            branch_first = branch_second
                        else:
                            branch_second, count = get_layer_uid(network_stack[i][j][k], count);
                            G.add_node(branch_second, pos=(column, branch_position));
                            branch_position += 1;
                            G.add_edge(branch_first, branch_second);
                            branch_first = branch_second;

                        branch_max_length = max(branch_position, branch_max_length)
                        if(k == len(network_stack[i][j])-1):
                            branch_end_points.append(branch_second);
                
                position = branch_max_length;
                position += 1;
                max_width += 2;

                sequential_second, count = get_layer_uid(network_stack[i][-1], count);

                G.add_node(sequential_second, pos=(2 + current_width, position));
                position += 1;
                for i in range(len(branch_end_points)):
                    G.add_edge(branch_end_points[i], sequential_second);
                sequential_first = sequential_second;
        else:
            sequential_second, count = get_layer_uid(network_stack[i], count);
            G.add_node(sequential_second, pos=(2+current_width, position))
            position += 1;
            G.add_edge(sequential_first, sequential_second);
            sequential_first = sequential_second;
    
    return G, count, sequential_second, position, max_width
    


#@TraceFunction(trace_args=True, trace_rv=False)
def debug_create_network(network_stack):
    '''
    Main function to visualize network when designing custom networks

    Args:
        network_stack (list): List of lists containing information on layers in the network
    
    Returns:
        neural network: The required complete network
    '''
    count = [];
    for i in range(len(names)):
        count.append(1);

    G=nx.DiGraph()
    G.add_node("Net", pos=(1,1))
    sequential_first = "data";
    #sequential_second, count = get_layer_uid(network_stack[0], count);

    count = [];
    for i in range(len(names)):
        count.append(1);

    position = 1;
    G.add_node(sequential_first, pos=(2,1))
    position += 1;


    max_width = 1;
    width = 0;
    for i in range(len(network_stack)):
        if(type(network_stack[i]) == list):
            is_block = True;

            if(type(network_stack[i][-1]) != list):
                if(network_stack[i][-1]["name"] in ["add", "concatenate"]):
                    is_block=False;

            if(is_block):
                G, count, sequential_second, position, _ =  debug_create_block(network_stack[i], count,
                                                                                G, sequential_first, position, 0)
                sequential_first = sequential_second
            else:
                branch_end_points = [];
                branch_max_length = 0;
                branches = [];
                branch_net = [];


                if(max_width < len(network_stack[i])-2):
                    max_width = len(network_stack[i])-2;
                width = 0;
                for j in range(len(network_stack[i])-1):
                    branch_first = sequential_first
                    branch_position = position
                    if(width > 0):
                        if(column == width):
                            column += 2;
                        else:
                            column = width;
                    else:
                        column = (j+1)*2;
                    for k in range(len(network_stack[i][j])):
                        if type(network_stack[i][j][k]) == list:
                            is_block2 = True;

                            if(type(network_stack[i][j][k][-1]) != list):
                                if(network_stack[i][j][k][-1]["name"] in ["add", "concatenate"]):
                                    is_block2=False;

                           

                            if(is_block2):
                                G, count, branch_second, branch_position, width = debug_create_block(network_stack[i][j][k], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j*2+width
                            else:
                                G, count, branch_second, branch_position, width = debug_create_block([network_stack[i][j][k]], 
                                                                                     count,
                                                                                     G, 
                                                                                     branch_first, 
                                                                                     branch_position, 
                                                                                     column-2) #j+k+width
                            branch_first = branch_second
                        else:
                            branch_second, count = get_layer_uid(network_stack[i][j][k], count);
                            G.add_node(branch_second, pos=(column, branch_position));
                            branch_position += 1;
                            G.add_edge(branch_first, branch_second);
                            branch_first = branch_second;

                        branch_max_length = max(branch_position, branch_max_length)
                        if(k == len(network_stack[i][j])-1):
                            branch_end_points.append(branch_second);
                    
                position = branch_max_length;
                position += 1;
                max_width += width

                sequential_second, count = get_layer_uid(network_stack[i][-1], count);

                G.add_node(sequential_second, pos=(2, position));
                position += 1;
                for i in range(len(branch_end_points)):
                    G.add_edge(branch_end_points[i], sequential_second);
                sequential_first = sequential_second;


        else:
            sequential_second, count = get_layer_uid(network_stack[i], count);
            G.add_node(sequential_second, pos=(2, position))
            position += 1;
            G.add_edge(sequential_first, sequential_second);
            sequential_first = sequential_second;

    max_width = max(max_width, width);
    if(max_width == 1):
        G.add_node("monk", pos=(3, position));
    else:
        G.add_node("monk", pos=(max_width + 3, position))
    pos = nx.get_node_attributes(G, 'pos')


    plt.figure(3, figsize=(16, 20 + position//6)) 
    nx.draw_networkx(G, pos, with_label=True, font_size=16, node_color="yellow", node_size=100)
    plt.savefig("graph.png");